/*
VDS.Web.Server is licensed under the MIT License

Copyright (c) 2010-2013 Robert Vesse

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Xml;
using System.Xml.Schema;
using VDS.Web.Consoles;
using VDS.Web.Handlers;
using VDS.Web.Logging;
using VDS.Web.Modules;

namespace VDS.Web.Configuration
{
    /// <summary>
    /// Static Helper class used for configuring the server
    /// </summary>
    public static class ServerConfiguration
    {
        /// <summary>
        /// Constant for the Configuration Namespace URI
        /// </summary>
        public const String NamespaceUri = "http://www.vdesign-studios.com/schemas/vds-web-server";

        /// <summary>
        /// Constant for the Schema Location that should be used in xsi:schemaLocation attributes
        /// </summary>
        public const String SchemaLocation = "http://www.vdesign-studios.com/schemas/vds-web-server http://www.vdesign-studios.com/schemas/vds-web-server.xsd";

        /// <summary>
        /// Constants for configuration elements
        /// </summary>
        public const String ElementServer = "server",
                            ElementAppSettings = "appSettings",
                            ElementAdd = "add",
                            ElementMimeTypes = "mimeTypes",
                            ElementMimeType = "mimeType",
                            ElementVirtualDirectories = "virtualDirs",
                            ElementVirtualDirectory = "virtualDir",
                            ElementModules = "modules",
                            ElementModule = "module",
                            ElementPreRequest = "prerequest",
                            ElementPreResponse = "preresponse",
                            ElementHandlers = "handlers",
                            ElementHandler = "handler",
                            ElementLoggers = "loggers",
                            ElementConsole = "console",
                            ElementFile = "file",
                            ElementProcess = "process",
                            ElementEvent = "event",
                            ElementMulticast = "multicast",
                            ElementExtensions = "extensions",
                            ElementExtension = "extension";

        /// <summary>
        /// Constants for configuration attributes
        /// </summary>
        public const String AttributeBaseDir = "basedir",
                            AttributeKey = "key",
                            AttributeValue = "value",
                            AttributePath = "path",
                            AttributeDirectory = "directory",
                            AttributeType = "type",
                            AttributeVerb = "verb",
                            AttributeFile = "file",
                            AttributeFormat = "format",
                            AttributeSource = "source",
                            AttributeLoadWith = "loadwith",
                            AttributeUseDefault = "usedefault",
                            AttributeBinary = "binary",
                            AttributeExtension = "extension";

        /// <summary>
        /// Configures the server
        /// </summary>
        /// <param name="server">Server</param>
        /// <param name="file">Configuration File</param>
        public static void Configure(HttpServer server, String file)
        {
            try
            {
                //Read in the Configuration File
                XmlDocument doc = new XmlDocument();
                doc.Load(file);

                //Run Configuration
                Configure(server, doc);
            }
            catch (HttpServerException)
            {
                //This catch exists since the try block may throw this error and if it does we don't want to wrap it which is what the general catch block will do
                throw;
            }
            catch (Exception ex)
            {
                throw new HttpServerException("Unable to create a HTTP Server instance since the Configuration File was invalid due to the following error: " + ex.Message, ex);
            }
        }

        /// <summary>
        /// Configures the server
        /// </summary>
        /// <param name="server">Server</param>
        /// <param name="config">XML Configuration</param>
        public static void Configure(HttpServer server, XmlDocument config)
        {
            try
            {
                if (config.DocumentElement.LocalName.Equals(ElementServer) && config.DocumentElement.NamespaceURI.Equals(NamespaceUri))
                {
                    if (config.DocumentElement.Attributes.Count > 0)
                    {
                        foreach (XmlAttribute attr in config.DocumentElement.Attributes)
                        {
                            if (!attr.NamespaceURI.Equals(NamespaceUri)) continue;

                            switch (attr.LocalName)
                            {
                                case AttributeBaseDir:
                                    String baseDir = attr.Value;
                                    if (baseDir.Equals(".")) baseDir = Environment.CurrentDirectory;
                                    if (!Directory.Exists(baseDir)) throw new HttpServerException("Configuration File specifies a base directory with the " + AttributeBaseDir + " attribute of the <" + ElementServer + "> element which is not a valid directory");
                                    server.BaseDirectory = baseDir;
                                    if (!server.BaseDirectory.EndsWith(new String(new char[] { Path.DirectorySeparatorChar }))) server.BaseDirectory += Path.DirectorySeparatorChar;
                                    break;
                            }
                        }
                    }

                    foreach (XmlNode child in config.DocumentElement.ChildNodes)
                    {
                        //Ignore non-element Nodes (e.g. comments)
                        if (child.NodeType != XmlNodeType.Element) continue;

                        if (!child.NamespaceURI.Equals(NamespaceUri)) continue;

                        switch (child.LocalName)
                        {
                            case ElementAppSettings:
                                ConfigureAppSettings(server, child);
                                break;

                            case ElementMimeTypes:
                                //MIME Type mappings
                                ConfigureMimeTypes(server, child);
                                break;

                            case ElementVirtualDirectories:
                                ConfigureVirtualDirectories(server, child);
                                break;

                            case ElementModules:
                                ConfigureModules(server, child);
                                break;

                            case ElementHandlers:
                                ConfigureHandlers(server, child);
                                break;

                            case ElementLoggers:
                                ConfigureLoggers(server, child);
                                break;

                            case ElementConsole:
                                if (child.ChildNodes.Count > 1)
                                {
                                    throw new HttpServerException("Configuration File contains more than one child element within the <" + ElementConsole + "> element, only a single console may be specified.  Use the <multicast> console to send console messages to multiple destinations");
                                }
                                else if (child.ChildNodes.Count == 1)
                                {
                                    server.Console = ConfigureConsole(server, child.FirstChild);
                                }
                                break;

                            default:
                                //Unknown node types are processed by an IConfigurationLoader specified by a @loadwith attribute
                                if (child.Attributes.GetNamedItem(AttributeLoadWith, NamespaceUri) != null)
                                {
                                    String loader = child.Attributes.GetNamedItem(AttributeLoadWith, NamespaceUri).Value;
                                    try
                                    {
                                        Object temp = Activator.CreateInstance(Type.GetType(loader));
                                        if (temp is IConfigurationLoader)
                                        {
                                            ((IConfigurationLoader)temp).Load(child, server);
                                        }
                                        else
                                        {
                                            throw new HttpServerException("Configuration File contains unexpected element <" + child.Name + "> whose loadwith attribute points to a type which does not implement the required IConfigurationLoader interface");
                                        }
                                    }
                                    catch (Exception ex)
                                    {
                                        throw new HttpServerException("Configuration File contains unexpected element <" + child.Name + "> whose loadwith attribute points to a type which could not be instantiated successfully or whose Load() method raised an exception", ex);
                                    }
                                }
                                else
                                {
                                    throw new HttpServerException("Configuration File contains unexpected element <" + child.Name + "> which does not have a loadwith attribute to specify a type which implement IConfigurationLoader and can load its configuration information");
                                }
                                break;
                        }
                    }
                }
                else
                {
                    throw new HttpServerException("Configuration File did not contain the required root element <" + ElementServer + ">");
                }
            }
            catch (HttpServerException)
            {
                //This catch exists since the try block may throw this error and if it does we don't want to wrap it which is what the general catch block will do
                throw;
            }
            catch (Exception ex)
            {
                throw new HttpServerException("Unable to create a HTTP Server instance since the Configuration File was invalid due to the following error: " + ex.Message, ex);
            }

            //If there was no <mimeTypes> section then use defaults
            if (server.MimeTypes == null) server.MimeTypes = new MimeTypeManager(true);
        }

        private static void ConfigureAppSettings(HttpServer server, XmlNode appSettings)
        {
            //Load App Settings
            foreach (XmlNode appSetting in appSettings.ChildNodes)
            {
                if (appSetting.NodeType != XmlNodeType.Element) continue;
                if (!appSetting.NamespaceURI.Equals(NamespaceUri)) throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementAppSettings + "> element is not in the required namespace");

                if (appSetting.LocalName.Equals(ElementAdd))
                {
                    String name = appSetting.Attributes.GetSafeNamedItem(AttributeKey, NamespaceUri);
                    if (name == null) throw new HttpServerException("Configuration File is invalid since an <" + ElementAdd + "> element in the <" + ElementAppSettings + "> section does not have a key attribute");
                    String value = appSetting.Attributes.GetSafeNamedItem(AttributeValue, NamespaceUri);

                    server.AppSettings[name] = value;
                }
                else
                {
                    throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementAppSettings + "> element is not an <" + ElementAdd + "> element");
                }
            }
        }

        private static void ConfigureMimeTypes(HttpServer server, XmlNode mimeTypes)
        {
            if (mimeTypes.NamespaceURI.Equals(NamespaceUri) && mimeTypes.LocalName.Equals(ElementMimeTypes))
            {
                //Create fresh manager
                MimeTypeManager manager;

                //Use Defaults?
                //Do this first so user specified ones can override these settings
                String useDefaults = mimeTypes.Attributes.GetSafeNamedItem(AttributeUseDefault, NamespaceUri);
                bool defaults = false;
                if (useDefaults != null)
                {
                    if (useDefaults.ToLower().Equals("true")) defaults = true;
                }
                manager = new MimeTypeManager(defaults);

                foreach (XmlNode mimeType in mimeTypes.ChildNodes)
                {
                    if (mimeType.NodeType != XmlNodeType.Element) continue;
                    if (!mimeType.NamespaceURI.Equals(NamespaceUri)) throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementMimeTypes + "> element is not in the required namespace");

                    if (mimeType.LocalName.Equals(ElementMimeType))
                    {
                        String ext = mimeType.Attributes.GetSafeNamedItem(AttributeExtension, NamespaceUri);
                        if (ext == null) throw new HttpServerException("Configuration File is invalid since a <mimeType> element does not have the required extension attribute");
                        String type = mimeType.Attributes.GetSafeNamedItem(AttributeType, NamespaceUri);
                        if (type == null) throw new HttpServerException("Configuration File is invalid since a <mimeTyoe> element does not have the required type attribute");
                        String bin = mimeType.Attributes.GetSafeNamedItem(AttributeBinary, NamespaceUri);
                        if (bin == null) bin = "false";
                        bin = bin.ToLower();

                        manager.AddMimeType(new MimeTypeMapping(ext, type, bin.Equals("true")));
                    }
                    else
                    {
                        throw new HttpServerException("Configuration File is invalid since the <" + ElementMimeTypes + "> element contains an unexpected <" + mimeType.Name + "> element");
                    }
                }

                server.MimeTypes = manager;
            }
            else
            {
                throw new HttpServerException("MimeTypeManager constructor expected to receive a <" + ElementMimeTypes + "> node but received a <" + mimeTypes.Name + "> node");
            }
        }

        private static void ConfigureVirtualDirectories(HttpServer server, XmlNode virtualDirs)
        {
            //Load Virtual Directories
            foreach (XmlNode vDir in virtualDirs.ChildNodes)
            {
                if (vDir.NodeType != XmlNodeType.Element) continue;
                if (!vDir.NamespaceURI.Equals(NamespaceUri)) throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementVirtualDirectories + "> element is not in the required namespace");

                if (vDir.LocalName.Equals(ElementVirtualDirectory))
                {
                    String path = vDir.Attributes.GetSafeNamedItem(AttributePath, NamespaceUri);
                    if (path == null) throw new HttpServerException("Configuration File is invalid since a <" + ElementVirtualDirectory + "> element does not have the required " + AttributePath + " attribute");
                    String dir = vDir.Attributes.GetSafeNamedItem(AttributeDirectory, NamespaceUri);
                    if (dir == null) throw new HttpServerException("Configuration File is invalid since a <" + ElementVirtualDirectory + "> element does not have the required " + AttributeBaseDir + " attribute");
                    if (!Directory.Exists(dir)) throw new HttpServerException("Configuration File is invalid since a <" + ElementVirtualDirectory + "> element specified a non-existent directory for its " + AttributeDirectory + " attribute");

                    server.VirtualDirectories.AddVirtualDirectory(path, dir);
                }
                else
                {
                    throw new HttpServerException("Configuration File is invalid since of the child elements of the <" + ElementVirtualDirectories + "> element is not a <" + ElementVirtualDirectory + "> element");
                }
            }
        }

        private static void ConfigureModules(HttpServer server, XmlNode modules)
        {
            //Load Modules
            foreach (XmlNode moduleTypeNode in modules.ChildNodes)
            {
                if (moduleTypeNode.NodeType != XmlNodeType.Element) continue;
                if (!moduleTypeNode.NamespaceURI.Equals(NamespaceUri)) throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementModules + "> element is not in the required namespace");

                String moduleType;
                switch (moduleTypeNode.LocalName)
                {
                    case ElementPreRequest:
                        foreach (XmlNode module in moduleTypeNode.ChildNodes)
                        {
                            if (module.NodeType != XmlNodeType.Element) continue;
                            if (!module.NamespaceURI.Equals(NamespaceUri) || !module.LocalName.Equals(ElementModule)) throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementPreRequest + "> element is not a <" + ElementModule + "> element");

                            moduleType = module.Attributes.GetSafeNamedItem(AttributeType, NamespaceUri);
                            if (moduleType == null) throw new HttpServerException("Configuration File is invalid since the a <" + ElementPreRequest + "> element does not have the required " + AttributeType + " attribute");

                            try
                            {
                                Object temp = Activator.CreateInstance(Type.GetType(moduleType));
                                if (temp is IHttpListenerModule)
                                {
                                    server.AddPreRequestModule((IHttpListenerModule)temp);
                                }
                                else
                                {
                                    throw new HttpServerException("Configuration File is invalid since it contains a <" + ElementPreRequest + "> element which specifies '" + moduleType + "' for its type attribute but this type does not implement the IHttpListenerModule interface");
                                }
                            }
                            catch (Exception ex)
                            {
                                throw new HttpServerException("Configuration File is invalid since it contains a <" + ElementPreRequest + "> element which specifies '" + moduleType + "' for its " + AttributeType + " attribute but this type and this type could not be instantiated successfully", ex);
                            }
                        }
                        break;
                    case ElementPreResponse:
                        foreach (XmlNode module in moduleTypeNode.ChildNodes)
                        {
                            if (module.NodeType != XmlNodeType.Element) continue;
                            if (!module.NamespaceURI.Equals(NamespaceUri) || !module.LocalName.Equals(ElementModule)) throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementPreRequest + "> element is not a <" + ElementModule + "> element");

                            moduleType = module.Attributes.GetSafeNamedItem(AttributeType, NamespaceUri);
                            if (moduleType == null) throw new HttpServerException("Configuration File is invalid since the a <" + ElementPreResponse + "> element does not have the required " + AttributeType + " attribute");

                            try
                            {
                                Object temp = Activator.CreateInstance(Type.GetType(moduleType));
                                if (temp is IHttpListenerModule)
                                {
                                    server.AddPreResponseModule((IHttpListenerModule)temp);
                                }
                                else
                                {
                                    throw new HttpServerException("Configuration File is invalid since it contains a <" + ElementPreResponse + "> element which specifies '" + moduleType + "' for its " + AttributeType + " attribute but this type does not implement the IHttpListenerModule interface");
                                }
                            }
                            catch (Exception ex)
                            {
                                throw new HttpServerException("Configuration File is invalid since it contains a <" + ElementPreResponse + "> element which specifies '" + moduleType + "' for its " + AttributeType + " attribute but this type and this type could not be instantiated successfully", ex);
                            }
                        }
                        break;
                    default:
                        throw new HttpServerException("Configuration File is invalid since the <" + ElementModules + "> element contains an unknown child element <" + moduleTypeNode.Name + ">");
                }
            }
        }

        private static void ConfigureHandlers(HttpServer server, XmlNode handlers)
        {
            //Load Handlers
            foreach (XmlNode handler in handlers.ChildNodes)
            {
                if (handler.NodeType != XmlNodeType.Element) continue;
                if (!handler.NamespaceURI.Equals(NamespaceUri)) throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementHandlers + "> element is not in the required namespace");

                if (handler.LocalName.Equals(ElementHandler))
                {
                    String verb = handler.Attributes.GetSafeNamedItem(AttributeVerb, NamespaceUri);
                    if (verb == null) verb = HttpRequestMapping.AllVerbs;
                    String path = handler.Attributes.GetSafeNamedItem(AttributePath, NamespaceUri);
                    if (path == null) path = HttpRequestMapping.AnyPath;
                    String type = handler.Attributes.GetSafeNamedItem(AttributeType, NamespaceUri);
                    if (type == null) throw new HttpServerException("Configuration File is invalid since one of the <" + ElementHandler + "> elements in the <" + ElementHandlers + "> section does not specify a " + AttributeType + " attribute");

                    Type t = Type.GetType(type);
                    if (t == null) throw new HttpServerException("Configuration File is invalid since the value '" + type + "' given for the " + AttributeType + " attribute of a <" + ElementHandler + "> element is not the type name of a valid type.  You may be missing an Assembly reference of failed to specify a Fully Qualified Type Name");
                    server.Handlers.AddMapping(new HttpRequestMapping(verb, path, t));
                }
                else
                {
                    throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementHandlers + "> element is not a <" + ElementHandler + "> element");
                }
            }
        }

        private static void ConfigureLoggers(HttpServer server, XmlNode loggers)
        {
            //Load Loggers
            foreach (XmlNode logger in loggers.ChildNodes)
            {
                if (logger.NodeType != XmlNodeType.Element) continue;
                if (!logger.NamespaceURI.Equals(NamespaceUri)) throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementLoggers + "> element is not in the required namespace");

                String logFormat;
                switch (logger.LocalName)
                {
                    case ElementFile:
                        String logFile = logger.Attributes.GetSafeNamedItem(AttributeFile, NamespaceUri);
                        if (logFile == null) throw new HttpServerException("Configuration File is invalid since a <" + ElementFile + "> element does not have the required " + AttributeFile + " attribute");
                        logFormat = logger.Attributes.GetSafeNamedItem(AttributeFormat, NamespaceUri);
                        if (logFormat == null) logFormat = "common";
                        logFormat = ApacheStyleLogger.GetLogFormat(logFormat);
                        server.AddLogger(new FileLogger(logFile, logFormat));
                        break;

                    case ElementProcess:
                        logFormat = logger.Attributes.GetSafeNamedItem(AttributeFormat, NamespaceUri);
                        if (logFormat == null) logFormat = "common";
                        logFormat = ApacheStyleLogger.GetLogFormat(logFormat);
                        server.AddLogger(new ConsoleLogger());
                        break;

                    case ElementEvent:
                        logFormat = logger.Attributes.GetSafeNamedItem(AttributeFormat, NamespaceUri);
                        if (logFormat == null) logFormat = "common";
                        logFormat = ApacheStyleLogger.GetLogFormat(logFormat);
                        String evtSource = logger.Attributes.GetSafeNamedItem(AttributeSource, NamespaceUri);
                        if (evtSource == null) evtSource = HttpServer.DefaultEventSource;
                        server.AddLogger(new EventLogger(evtSource));
                        break;

                    case ElementExtension:
                        String loader = logger.Attributes.GetSafeNamedItem(AttributeLoadWith, NamespaceUri);
                        if (loader == null) throw new HttpServerException("Configuration File is invalid since a <" + ElementExtension + "> element does not have the required " + AttributeLoadWith + " attribute");
                        try
                        {
                            Type t = Type.GetType(loader);
                            if (t == null) throw new HttpServerException("Configuration File is invalid since the value '" + loader + "' given for the " + AttributeType + " attribute of a logger extension is not the type name of a valid type.  You may be missing an Assembly reference of failed to specify a Fully Qualified Type Name");
                            Object temp = Activator.CreateInstance(t);
                            if (temp is IConfigurationLoader)
                            {
                                ((IConfigurationLoader)temp).Load(logger, server);
                            }
                            else
                            {
                                throw new HttpServerException("Configuration File contains unexpected logger extension whose " + AttributeLoadWith + " attribute points to a type which does not implement the required IConfigurationLoader interface");
                            }
                        }
                        catch (HttpServerException)
                        {
                            //Don't re-wrap these
                            throw;
                        }
                        catch (Exception ex)
                        {
                            throw new HttpServerException("Configuration File contains unexpected logger extension whose " + AttributeLoadWith + " attribute points to a type which could not be instantiated successfully or whose Load() method raised an exception", ex);
                        }
                        break;

                    default:
                        throw new HttpServerException("Configuration File contains unexpected child element <" + logger.Name + "> under the <" + ElementLoggers + "> element");
                }
            }
        }

        private static IServerConsole ConfigureConsole(HttpServer server, XmlNode console)
        {
            if (console.NodeType != XmlNodeType.Element) return null;
            if (!console.NamespaceURI.Equals(NamespaceUri)) throw new HttpServerException("Configuration File is invalid since one of the child elements of the <" + ElementConsole + "> element is not in the required namespace");

            //Load Console
            switch (console.LocalName)
            {
                case ElementFile:
                    String consoleFile = console.Attributes.GetSafeNamedItem(AttributeFile, NamespaceUri);
                    if (consoleFile == null) throw new HttpServerException("Configuration File is invalid since a <file> console element does not have the required file attribute");
                    return new FileConsole(consoleFile);

                case ElementProcess:
                    return new ProcessConsole();

                case ElementEvent:
                    String evtSource = console.Attributes.GetSafeNamedItem(AttributeSource, NamespaceUri);
                    if (evtSource == null) evtSource = HttpServer.DefaultEventSource;
                    return new EventConsole(evtSource);

                case ElementMulticast:
                    MulticastConsole mcast = new MulticastConsole();
                    foreach (XmlNode dest in console.ChildNodes)
                    {
                        mcast.AddConsole(ConfigureConsole(server, dest));
                    }
                    return mcast;

                case ElementExtension:
                    String loader = console.Attributes.GetSafeNamedItem(AttributeLoadWith, NamespaceUri);
                    if (loader == null) throw new HttpServerException("Configuration File is invalid since a <" + ElementExtension + "> element does not have the required " + AttributeLoadWith + " attribute");
                    try
                    {
                        Type t = Type.GetType(loader);
                        if (t == null) throw new HttpServerException("Configuration File is invalid since the value '" + loader + "' given for the " + AttributeType + " attribute of a console extension is not the type name of a valid type.  You may be missing an Assembly reference of failed to specify a Fully Qualified Type Name");
                        Object temp = Activator.CreateInstance(t);
                        if (temp is IConfigurationLoader)
                        {
                            ((IConfigurationLoader)temp).Load(console, server);
                            return server.Console;
                        }
                        else
                        {
                            throw new HttpServerException("Configuration File contains unexpected console extension whose " + AttributeLoadWith + " attribute points to a type which does not implement the required IConfigurationLoader interface");
                        }
                    }
                    catch (HttpServerException)
                    {
                        //Don't re-wrap
                        throw;
                    }
                    catch (Exception ex)
                    {
                        throw new HttpServerException("Configuration File contains unexpected console extension whose " + AttributeLoadWith + " attribute points to a type which could not be instantiated successfully or whose Load() method raised an exception", ex);
                    }

                default:
                    throw new HttpServerException("Configuration File contains unexpected child element <" + console.Name + "> under the <" + ElementConsole + "> element");
            }

        }

        /// <summary>
        /// Validates whether a given XML configuration file according to the schema
        /// </summary>
        /// <param name="file">File</param>
        public static void Validate(String file)
        {
            if (File.Exists(file))
            {
                using (Stream stream = File.OpenRead(file))
                {
                    Validate(stream);
                    stream.Close();
                }
            }
            else
            {
                throw new XmlException("Given configuration file does not exist");
            }
        }

        /// <summary>
        /// Validates whether a given XML configuration file is valid according to the schema
        /// </summary>
        /// <param name="input">Input Stream</param>
        public static void Validate(Stream input)
        {
            // configure the xmlreader validation to use inline schema.
            XmlReaderSettings config = new XmlReaderSettings();
            config.ValidationType = ValidationType.Schema;
            config.ValidationFlags |= XmlSchemaValidationFlags.ReportValidationWarnings;
            config.ValidationFlags |= XmlSchemaValidationFlags.ProcessInlineSchema;
            config.ValidationFlags |= XmlSchemaValidationFlags.ProcessSchemaLocation;
            config.ValidationEventHandler += (sender, args) => { throw args.Exception; };

            // Get the XmlReader object with the configured settings.
            XmlReader reader = XmlReader.Create(input, config);

            // Parsing the file will cause the validation to occur.
            while (reader.Read()) ;
        }
    }
}
