/*
VDS.Web.Server is licensed under the MIT License

Copyright (c) 2010-2013 Robert Vesse

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace VDS.Web.Consoles
{
    /// <summary>
    /// Abstract base class for consoles which prettifies the output with timestamps and type prefixes
    /// </summary>
    public abstract class BaseConsole
        : IServerConsole
    {
        /// <summary>
        /// Constants for prefixes applied to console messages
        /// </summary>
        private const String ErrorPrefix = "ERROR ",
                             WarnPrefix = "WARN  ",
                             InformationPrefix = "INFO  ",
                             DebugPrefix = "DEBUG ";

        public const String DefaultTimestampFormat = "dd/MM/yyyy HH:mm:ss.fff K";

        private bool _useTimestamps = true;
        private String _timestampFormat = DefaultTimestampFormat;

        public BaseConsole()
            : this(true) { }

        public BaseConsole(bool useTimestamps)
            : this(useTimestamps, DefaultTimestampFormat) { }

        public BaseConsole(bool useTimestamps, String timestampFormat)
        {
            if (useTimestamps && timestampFormat == null) throw new ArgumentNullException("timestampFormat", "Timestamp Format cannot be null if useTimestamps is true");
            this._useTimestamps = true;
            this._timestampFormat = timestampFormat;
        }

        private String GetPrefix(String type)
        {
            String prefix = String.Empty;
            if (this._useTimestamps) 
            {
                prefix = DateTime.Now.ToString(this._timestampFormat);
                prefix += " ";
            }
            prefix += type;
            return prefix;
        }

        public void Information(string message)
        {
            this.PrintInformation(this.GetPrefix(InformationPrefix) + message);
        }

        protected abstract void PrintInformation(String message);

        public void Warning(string message)
        {
            this.PrintWarning(this.GetPrefix(WarnPrefix) + message);
        }

        protected abstract void PrintWarning(String message);

        public void Warning(string message, Exception e)
        {
            this.PrintWarning(this.GetPrefix(WarnPrefix) + message, e);
        }

        protected abstract void PrintWarning(String message, Exception e);

        public void Error(string message)
        {
            this.PrintError(this.GetPrefix(ErrorPrefix) + message);
        }

        protected abstract void PrintError(String message);

        public void Error(string message, Exception e)
        {
            this.PrintError(this.GetPrefix(ErrorPrefix) + message);
        }

        protected abstract void PrintError(String message, Exception e);

        public void Error(Exception e)
        {
            this.Error(e.Message, e);
        }

        /// <summary>
        /// Helper method which returns the full stack trace for an exception
        /// </summary>
        /// <param name="e">Exception</param>
        /// <returns>Full Stack Trace</returns>
        protected String GetFullStackTrace(Exception e)
        {
            if (e == null) return String.Empty;

            StringBuilder builder = new StringBuilder();
            do
            {
                builder.AppendLine(e.Message);
                builder.AppendLine(e.StackTrace);

                if (e.InnerException != null) builder.Append("Caused by: ");
                e = e.InnerException;
            } while (e != null);
            return builder.ToString();
        }

        public virtual void PreStateChange(HttpServer server, ServerState target)
        {
            //Do nothing
        }

        public void PostStateChange(HttpServer server)
        {
            //Do nothing
        }

        public abstract void Dispose();
    }
}
