/*
VDS.Web.Server is licensed under the MIT License

Copyright (c) 2010-2013 Robert Vesse

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net;
using System.Threading;
using System.Xml;
using VDS.Web.Configuration;
using VDS.Web.Consoles;
using VDS.Web.Handlers;
using VDS.Web.Logging;
using VDS.Web.Modules;

namespace VDS.Web
{
    /// <summary>
    /// A simple but extensible HTTP Server based on a <see cref="HttpListener">HttpListener</see>
    /// </summary>
    public class HttpServer
        : IDisposable
    {
        private HttpListener _listener;
        private IHttpListenerHandlerCollection _handlers;
        private List<IHttpListenerModule> _preRequestModules = new List<IHttpListenerModule>();
        private List<IHttpListenerModule> _preResponseModules = new List<IHttpListenerModule>();
        private VirtualDirectoryManager _virtualDirs = new VirtualDirectoryManager();
        private MimeTypeManager _mimeTypes;
        private IServerConsole _console = new ProcessConsole();

        private Thread _serverThread;
        private bool _shouldTerminate = false;
        private HandleRequestDelegate _delegate;
        private bool _disposed = false;

        private int _port = DefaultPort;
        private String _host = DefaultHost;
        private String _baseDirectory = null;

        private Context _context = new Context(), _settings = new Context();
        private List<IHttpLogger> _loggers = new List<IHttpLogger>();

        /// <summary>
        /// Constant for using accepting requests to the port from any host name
        /// </summary>
        public const String AnyHost = "*";
        /// <summary>
        /// Constant for the Default Host used if none is specified (localhost)
        /// </summary>
        public const String DefaultHost = "localhost";
        /// <summary>
        /// Constant for the Default Port used if none is specified (80)
        /// </summary>
        public const int DefaultPort = 80;
        /// <summary>
        /// Constant for the Default Configuration File used
        /// </summary>
        public const String DefaultConfigurationFile = "server.config";
        /// <summary>
        /// Constant for default event source for System Event Log based logging
        /// </summary>
        public const String DefaultEventSource = "VDS.Web.Server.HttpServer";

        #region Constructors

        #region Explicit Constructors

        /// <summary>
        /// Creates a new Server with the given host
        /// </summary>
        /// <param name="host">Host</param>
        /// <remarks>
        /// Must be explicitly started
        /// </remarks>
        public HttpServer(String host)
            : this(host, new HttpListenerHandlerCollection(), false) { }

        /// <summary>
        /// Creates a new Server with the given host and optionally starts the server
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="autostart">Whether to auto-start</param>
        public HttpServer(String host, bool autostart)
            : this(host, new HttpListenerHandlerCollection(), autostart) { }

        /// <summary>
        /// Creates a new Server with the given handlers
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="handlers">Handlers</param>
        /// <remarks>
        /// Must be explicitly started
        /// </remarks>
        public HttpServer(String host, HttpListenerHandlerCollection handlers)
            : this(host, handlers, false) { }

        /// <summary>
        /// Creates a new Server on the given port
        /// </summary>
        /// <param name="port">Port</param>
        /// <remarks>
        /// Must be explicitly started
        /// </remarks>
        public HttpServer(int port)
            : this(port, new HttpListenerHandlerCollection()) { }

        /// <summary>
        /// Creates a new Server on the given port and optionally starts the server
        /// </summary>
        /// <param name="port">Port</param>
        /// <param name="autostart">Whether to auto-start</param>
        public HttpServer(int port, bool autostart)
            : this(port, new HttpListenerHandlerCollection(), autostart) { }

        /// <summary>
        /// Creates a new Server on the given port with the given handlers
        /// </summary>
        /// <param name="port">Port</param>
        /// <param name="handlers">Handlers</param>
        /// <remarks>
        /// Must be explicitly started
        /// </remarks>
        public HttpServer(int port, IHttpListenerHandlerCollection handlers)
            : this(port, handlers, false) { }

        /// <summary>
        /// Creates a new Server on the given host with the given handlers
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="handlers">Handlers</param>
        /// <param name="autostart">Whether to auto-start</param>
        public HttpServer(String host, IHttpListenerHandlerCollection handlers, bool autostart)
            : this(host, DefaultPort, handlers, autostart) { }

        /// <summary>
        /// Creates a new Server on the given port with the given handlers
        /// </summary>
        /// <param name="port">Port</param>
        /// <param name="handlers">Handlers</param>
        /// <param name="autostart">Whether to auto-start</param>
        public HttpServer(int port, IHttpListenerHandlerCollection handlers, bool autostart)
            : this(DefaultHost, port, handlers, autostart) { }

        /// <summary>
        /// Creates a new Server on the given host and port
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="port">Port</param>
        /// <remarks>
        /// Must be explicitly started
        /// </remarks>
        public HttpServer(String host, int port)
            : this(host, port, new HttpListenerHandlerCollection(), false) { }

        /// <summary>
        /// Creates a new Server on the given host and port and optionally starts the server
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="port">Port</param>
        /// <param name="autostart">Whether to auto-start</param>
        public HttpServer(String host, int port, bool autostart)
            : this(host, port, new HttpListenerHandlerCollection(), autostart) { }

        /// <summary>
        /// Creates a new Server on the given host and port with the given handlers
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="port">Port</param>
        /// <param name="handlers">Handlers</param>
        /// <remarks>
        /// Must be explicitly started
        /// </remarks>
        public HttpServer(String host, int port, IHttpListenerHandlerCollection handlers)
            : this(host, port, handlers, false) { }

        /// <summary>
        /// Creates a new Server on the given host and port with the given handlers and optionally starts the server
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="port">Port</param>
        /// <param name="handlers">Handlers</param>
        /// <param name="autostart">Whether to auto-start</param>
        public HttpServer(String host, int port, IHttpListenerHandlerCollection handlers, bool autostart)
        {
            if (host == null) throw new ArgumentNullException("host", "Cannot specify a null host, to specify any host please use the constant HttpServer.AnyHost");
            if (handlers == null) throw new ArgumentNullException("handlers", "Cannot specify a null Handlers Collection");
            if (!HttpListener.IsSupported) throw new PlatformNotSupportedException("HttpServer is not able to run on your Platform as HttpListener is not supported");

            this._handlers = handlers;
            this._mimeTypes = new MimeTypeManager(true);
            this._preResponseModules.Add(new LoggingModule());
            this.Initialise(host, port);

            //Auto-start if specified
            if (autostart) this.Start();
        }

        #endregion

        #region Using Configuration File Constructors

        /// <summary>
        /// Creates a new Server on the given host and port using a specified configuration file and optionally starts the server
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="port">Port</param>
        /// <param name="configFile">Configuration File</param>
        /// <param name="autostart">Whether to auto-start</param>
        public HttpServer(String host, int port, String configFile, bool autostart)
        {
            if (host == null) throw new ArgumentNullException("host", "Cannot specify a null host, to specify any host please use the constant HttpServer.AnyHost");
            if (!HttpListener.IsSupported) throw new PlatformNotSupportedException("HttpServer is not able to run on your Platform as HttpListener is not supported");
            if (!File.Exists(configFile)) throw new ArgumentException("Cannot specify a non-existent Configuration File for the Server", "configFile");
            
            this._handlers = new HttpListenerHandlerCollection();

            ServerConfiguration.Configure(this, configFile);

            this.Initialise(host, port);

            //Auto-start if specified
            if (autostart) this.Start();
        }

        /// <summary>
        /// Creates a new Server on the given host and port using a specified configuration file
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="port">Port</param>
        /// <param name="configFile">Configuration File</param>
        /// <remarks>
        /// Must be explicitly started
        /// </remarks>
        public HttpServer(String host, int port, String configFile)
            : this(host, port, configFile, false) { }

        /// <summary>
        /// Creates a new Server on the given host using a specified configuration file and optionally starts the server
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="configFile">Configuration File</param>
        /// <param name="autostart">Whether to auto-start</param>
        public HttpServer(String host, String configFile, bool autostart)
            : this(host, DefaultPort, configFile, autostart) { }

        /// <summary>
        /// Creates a new Server on the given host using a specified configuration file
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="configFile">Configuration File</param>
        /// <remarks>
        /// Must be explicitly started
        /// </remarks>
        public HttpServer(String host, String configFile)
            : this(host, DefaultPort, configFile, false) { }

        /// <summary>
        /// Creates a new Server on the given port using a specified configuration file and optionally starts the server
        /// </summary>
        /// <param name="port">Port</param>
        /// <param name="configFile">Configuration File</param>
        /// <param name="autostart">Whether to auto-start</param>
        public HttpServer(int port, String configFile, bool autostart)
            : this(DefaultHost, port, configFile, autostart) { }

        /// <summary>
        /// Creates a new Server on the given port using a specified configuration file
        /// </summary>
        /// <param name="port">Port</param>
        /// <param name="configFile">Configuration File</param>
        /// <remarks>
        /// Must be explicitly started
        /// </remarks>
        public HttpServer(int port, String configFile)
            : this(DefaultHost, port, configFile, false) { }

        #endregion

        /// <summary>
        /// Destructor which ensures the server is cleaned up
        /// </summary>
        ~HttpServer()
        {
            this.DisposeInternal(false);
        }

        /// <summary>
        /// Initialisation of the Server
        /// </summary>
        /// <param name="host">Host</param>
        /// <param name="port">Port</param>
        private void Initialise(String host, int port)
        {
            this.State = ServerState.Creating;

            //Tell stuff that we're being created
            this.PreStateChange(ServerState.Created);

            //Set up HTTP Listener
            this._port = port;
            this._host = host;
            this._listener = new HttpListener();
            this._listener.Prefixes.Add("http://" + host + ":" + port + "/");
            this._listener.IgnoreWriteExceptions = true;

            //Set up Request Handling Thread
            this._delegate = new HandleRequestDelegate(this.HandleRequest);
            this._serverThread = new Thread(new ThreadStart(this.Run));
            this._serverThread.Start();

            //We're now created
            this.State = ServerState.Created;
            this.PostStateChange();
        }

        #endregion

        #region Server Properties

        /// <summary>
        /// Gets whether the Server is running
        /// </summary>
        public bool IsRunning
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this.State == ServerState.Running;
            }
        }

        /// <summary>
        /// Gets the Port the Server is using
        /// </summary>
        public int Port
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._port;
            }
        }

        /// <summary>
        /// Gets the Host the Server is using
        /// </summary>
        public String Host
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._host;
            }
        }

        /// <summary>
        /// Gets the Base Directory used for static content
        /// </summary>
        public String BaseDirectory
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._baseDirectory;
            }
            set
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                this._baseDirectory = value;
            }
        }

        /// <summary>
        /// Gets the application settings
        /// </summary>
        public Context AppSettings
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._settings;
            }
        }

        /// <summary>
        /// Gets the context of the Server
        /// </summary>
        /// <remarks>
        /// The context is a name value collection which persists only while the server is running, if you need to persist beyond that use the <see cref="HttpServer.AppSettings"/> instead
        /// </remarks>
        public Context Context
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._context;
            }
        }

        /// <summary>
        /// Gets/Sets the console which is used to log non-HTTP request specific messages, all HTTP request logging is handled by the <see cref="IHttpLogger" /> infrastructure.
        /// </summary>
        /// <remarks>
        /// Setting the console to null means that informational messages will be silently discarded.
        /// </remarks>
        public IServerConsole Console
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._console;
            }
            set
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                if (value != null)
                {
                    this._console = value;
                }
                else
                {
                    this._console = new NullConsole();
                }
            }
        }

        /// <summary>
        /// Gets the Handlers for the Server
        /// </summary>
        public IHttpListenerHandlerCollection Handlers
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._handlers;
            }
        }

        /// <summary>
        /// Gets the Loggers for the Server
        /// </summary>
        public IEnumerable<IHttpLogger> Loggers
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._loggers;
            }
        }

        /// <summary>
        /// Gets/Sets the MIME type manager for the server
        /// </summary>
        public MimeTypeManager MimeTypes
        {
            get
            {
                return this._mimeTypes;
            }
            set
            {
                if (value != null)
                {
                    this._mimeTypes = value;
                }
            }
        }

        /// <summary>
        /// Gets the pre-request modules for the Server
        /// </summary>
        public IEnumerable<IHttpListenerModule> PreRequestModules
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._preRequestModules;
            }
        }

        /// <summary>
        /// Gets the pre-response modules for the Server
        /// </summary>
        public IEnumerable<IHttpListenerModule> PreResponseModules
        {
            get
            {
                if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
                return this._preResponseModules;
            }
        }

        /// <summary>
        /// Gets the server state
        /// </summary>
        public ServerState State
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets/Sets the virtual directory manager
        /// </summary>
        public VirtualDirectoryManager VirtualDirectories
        {
            get
            {
                return this._virtualDirs;
            }
            set
            {
                if (value == null)
                {
                    this._virtualDirs = new VirtualDirectoryManager();
                }
                else
                {
                    this._virtualDirs = value;
                }
            }
        }

        #endregion

        #region Server Configuration

        /// <summary>
        /// Adds a virtual directory to the Server
        /// </summary>
        /// <param name="path">Virtual Path</param>
        /// <param name="directory">Physical Directory</param>
        [Obsolete("You can now access the Virtual Directory manager via the VirtualDirectories property, please call methods on that directly", true)]
        public void AddVirtualDirectory(String path, String directory)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            if (this._virtualDirs == null) this._virtualDirs = new VirtualDirectoryManager();
            this._virtualDirs.AddVirtualDirectory(path, directory);
        }

        /// <summary>
        /// Adds a MIME Type to the Server
        /// </summary>
        /// <param name="ext">File Extension</param>
        /// <param name="mimeType">MIME Type</param>
        /// <param name="binary">Whether the type should be treated as binary content</param>
        [Obsolete("You can now access the MIME type manager via the MimeTypes property, please call methods on that directly", true)]
        public void AddMimeType(String ext, String mimeType, bool binary)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._mimeTypes.AddMimeType(ext, mimeType, binary);
        }

        /// <summary>
        /// Adds a MIME Type to the Server
        /// </summary>
        /// <param name="ext">File Extension</param>
        /// <param name="mimeType">MIME Type</param>
        [Obsolete("You can now access the MIME type manager via the MimeTypes property, please call methods on that directly", true)]
        public void AddMimeType(String ext, String mimeType)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._mimeTypes.AddMimeType(ext, mimeType);
        }

        /// <summary>
        /// Adds a pre-request module to the Server
        /// </summary>
        /// <param name="module">Module</param>
        public void AddPreRequestModule(IHttpListenerModule module)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._preRequestModules.Add(module);
        }

        /// <summary>
        /// Adds a pre-response module to the Server
        /// </summary>
        /// <param name="module">Module</param>
        public void AddPreResponseModule(IHttpListenerModule module)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._preResponseModules.Add(module);
        }

        /// <summary>
        /// Inserts a pre-request module to the Server
        /// </summary>
        /// <param name="module">Module</param>
        /// <param name="insertAt">Index to insert at</param>
        public void InsertPreRequestModule(IHttpListenerModule module, int insertAt)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._preRequestModules.Insert(insertAt, module);
        }

        /// <summary>
        /// Inserts a pre-response module to the Server
        /// </summary>
        /// <param name="module">Module</param>
        /// <param name="insertAt">Index to insert at</param>
        public void InsertPreResponseModule(IHttpListenerModule module, int insertAt)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._preResponseModules.Insert(insertAt, module);
        }

        /// <summary>
        /// Removes a pre-request module to the Server
        /// </summary>
        /// <param name="module">Module</param>
        public void RemovePreRequestModule(IHttpListenerModule module)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._preRequestModules.Remove(module);
        }

        /// <summary>
        /// Removes a pre-response module to the Server
        /// </summary>
        /// <param name="module">Module</param>
        public void RemovePreResponseModule(IHttpListenerModule module)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._preResponseModules.Remove(module);
        }

        /// <summary>
        /// Removes all pre-request modules
        /// </summary>
        public void ClearPreRequestModules()
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._preRequestModules.Clear();
        }

        /// <summary>
        /// Removes all pre-response modules
        /// </summary>
        public void ClearPreResponseModules()
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._preResponseModules.Clear();
        }

        /// <summary>
        /// Adds a Logger
        /// </summary>
        /// <param name="logger">Logger</param>
        public void AddLogger(IHttpLogger logger)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._loggers.Add(logger);
        }

        /// <summary>
        /// Inserts a Logger
        /// </summary>
        /// <param name="logger">Logger</param>
        /// <param name="insertAt">Index to insert at</param>
        public void InsertLogger(IHttpLogger logger, int insertAt)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._loggers.Insert(insertAt, logger);
        }

        /// <summary>
        /// Removes a Logger
        /// </summary>
        /// <param name="logger">Logger</param>
        public void RemoveLogger(IHttpLogger logger)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._loggers.Remove(logger);
        }

        /// <summary>
        /// Removes all loggers
        /// </summary>
        public void ClearLoggers()
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");
            this._loggers.Clear();
        }

        #endregion

        #region Server Control

        /// <summary>
        /// Starts the Server
        /// </summary>
        public void Start()
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");

            try
            {
                Monitor.Enter(this);

                //Only start if not currently running
                if (this.State != ServerState.Running)
                {
                    //Tell stuff that we are starting
                    this.PreStateChange(ServerState.Starting);
                    this._console.Information("HttpServer Starting on " + this._host + ":" + this._port);
                    this.State = ServerState.Starting;
                    this.PostStateChange();

                    //Start the server
                    this.PreStateChange(ServerState.Running);
                    this._listener.Start();
                    while (!this._listener.IsListening)
                    {
                        //Wait for listener to start
                    }
                    this._console.Information("HttpServer Started");
                    this.State = ServerState.Running;
                    this.PostStateChange();
                }
                else
                {
                    throw new HttpServerException("Server is already running, calling Start() is not valid at this time");
                }
            }
            catch (Exception e)
            {
                this.LogError("An error occurred while trying to start the server", e);
                throw;
            }
            finally
            {
                Monitor.Exit(this);
            }
        }

        /// <summary>
        /// Restarts the Server
        /// </summary>
        public void Restart()
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");

            this._console.Warning("HttpServer Restarting...");
            try
            {
                Monitor.Enter(this);
                this.Stop();
                this.Start();
            }
            finally
            {
                Monitor.Exit(this);
            }
        }

        /// <summary>
        /// Stops the Server
        /// </summary>
        public void Stop()
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");

            try
            {
                Monitor.Enter(this);

                //Only Stop if we are actually running
                if (this.State == ServerState.Running)
                {
                    //Tell stuff that we are stopping
                    this._console.Information("HttpServer Stopping...");
                    this.PreStateChange(ServerState.Stopping);
                    this.State = ServerState.Stopping;
                    this.PostStateChange();

                    //Perform the actual stop
                    this.PreStateChange(ServerState.Stopped);
                    this._listener.Stop();
                    this.State = ServerState.Stopped;
                    this.PostStateChange();
                    this._console.Information("HttpServer Stopped");

                    //Post-stop clean up
                    this._context.Clear();
                }
                else
                {
                    throw new HttpServerException("Server is not currently running, calling Stop() is not valid at this time");
                }
            }
            catch (Exception e)
            {
                this.LogError("An error occurred while trying to stop the server", e);
                throw;
            }
            finally
            {
                Monitor.Exit(this);
            }
        }

        /// <summary>
        /// Shuts down the server and optionally stops the process in which it is running
        /// </summary>
        /// <param name="stopProcess">Whether to stop the containing process</param>
        /// <param name="force">Whether to force the containing process to stop by killing it</param>
        /// <remarks>
        /// <para>
        /// After a Server has been shut down it's methods and properties cannot be accessed
        /// </para>
        /// </remarks>
        public void Shutdown(bool stopProcess, bool force)
        {
            if (this._disposed) throw new ObjectDisposedException("Cannot access the properties/methods of a HttpServer instance after it has been disposed of");

            try
            {
                Monitor.Enter(this);

                this._console.Information("HttpServer shutting down...");
                this.Dispose();
                this._console.Information("HttpServer shut down");

                if (stopProcess)
                {
                    Process p = System.Diagnostics.Process.GetCurrentProcess();
                    this._console.Information("HttpServer " + (force ? "killing" : "closing") + " containing process...");
                    if (force)
                    {
                        p.Kill();
                    }
                    else
                    {
                        p.Close();
                    }
                }
            }
            finally
            {
                Monitor.Exit(this);
            }
        }

        /// <summary>
        /// Helper method which invokes <see cref="ILifecycleAwareComponent.PreStateChange(HttpServer,ServerState)"/> on relevant server components
        /// </summary>
        /// <param name="target">Target state</param>
        private void PreStateChange(ServerState target)
        {
            try
            {
                this._handlers.PreStateChange(this, target);
                foreach (IHttpListenerModule module in this._preRequestModules)
                {
                    try
                    {
                        module.PreStateChange(this, target);
                    }
                    catch (Exception mEx)
                    {
                        this.LogError("Error sending pre-state change notification to module " + module.GetType().FullName, mEx);
                    }
                }
                foreach (IHttpListenerModule module in this._preResponseModules)
                {
                    try
                    {
                        module.PreStateChange(this, target);
                    }
                    catch (Exception mEx)
                    {
                        this.LogError("Error sending pre-state change notification to module " + module.GetType().FullName, mEx);
                    }
                }
                this._console.PreStateChange(this, target);
            }
            catch (Exception e)
            {
                this.LogError("Error sending pre-state change notifications", e);
                throw;
            }
        }

        /// <summary>
        /// Helper method which invokes <see cref="ILifecycleAwareComponent.PostStateChange(HttpServer)"/> on relevant server components
        /// </summary>
        private void PostStateChange()
        {
            try
            {
                this._handlers.PostStateChange(this);
                foreach (IHttpListenerModule module in this._preRequestModules)
                {
                    try
                    {
                        module.PostStateChange(this);
                    }
                    catch (Exception mEx)
                    {
                        this.LogError("Error sending post-state change notification to module " + module.GetType().FullName, mEx);
                    }
                }
                foreach (IHttpListenerModule module in this._preResponseModules)
                {
                    try
                    {
                        module.PostStateChange(this);
                    }
                    catch (Exception mEx)
                    {
                        this.LogError("Error sending post-state change notification to module " + module.GetType().FullName, mEx);
                    }
                }
                this._console.PostStateChange(this);
            }
            catch (Exception e)
            {
                this.LogError("Error sending post-state change notifications", e);
                throw;
            }
        }

        #endregion

        #region Request Processing

        /// <summary>
        /// Actual method which implements the main server process
        /// </summary>
        private void Run()
        {
            while (true)
            {
                if (this._disposed || this._shouldTerminate) return;

                if (this._listener.IsListening)
                {
                    try
                    {
                        HttpListenerContext context = this._listener.GetContext();

                        //Hand it off to be processed asynchronously
                        this._delegate.BeginInvoke(context, new AsyncCallback(this.EndRequest), null);
                    }
                    catch (Exception ex)
                    {
                        if (this._listener.IsListening) this.LogError(ex);
                    }
                }
                else
                {
                    //Wait until we're asked to listen again
                    Thread.Sleep(50);
                }
            }
        }

        /// <summary>
        /// Delegate for handing off request processing to be asynchronous
        /// </summary>
        /// <param name="context">Listener Context</param>
        /// <returns></returns>
        private delegate HttpServerContext HandleRequestDelegate(HttpListenerContext context);

        /// <summary>
        /// Begins request handling
        /// </summary>
        /// <param name="context">Listener Context</param>
        /// <returns></returns>
        private HttpServerContext HandleRequest(HttpListenerContext context)
        {
            IHttpListenerHandler handler;
            HttpServerContext serverContext = new HttpServerContext(this, context);
            try
            {
                bool skipHandling = this.ApplyPreRequestModules(serverContext);
                if (!skipHandling)
                {
                    handler = this._handlers.GetHandler(serverContext);
                    try
                    {
                        handler.ProcessRequest(serverContext);
                    }
                    catch (HttpServerException)
                    {
                        throw;
                    }
                    catch (Exception ex)
                    {
                        throw new HttpServerException("Unexpected error from Handler " + handler.GetType().FullName, ex);
                    }
                }
            }
            catch (NoHandlerException noHandlerEx)
            {
                this.LogError(noHandlerEx);
                context.Response.StatusCode = (int)HttpStatusCode.NotFound;
            }
            catch (HttpServerException serverEx)
            {
                this.LogError(serverEx);
                context.Response.StatusCode = (int)HttpStatusCode.InternalServerError;

                //Write out the error information to the client
                using (StreamWriter writer = new StreamWriter(context.Response.OutputStream))
                {
                    Exception e = serverEx;
                    do
                    {
                        writer.WriteLine(e.Message);
                        writer.WriteLine(e.StackTrace);

                        if (e.InnerException != null) writer.Write("Caused by: ");
                        e = e.InnerException;
                    } while (e != null);
                }
            }
            catch (Exception ex)
            {
                this.LogError("Unexpected Error in HttpServer.HandleRequest()", ex);
            }

            return serverContext;
        }

        /// <summary>
        /// Ends request handling
        /// </summary>
        /// <param name="result">Async Result</param>
        private void EndRequest(IAsyncResult result)
        {
            HttpServerContext context = null;
            try
            {
                context = this._delegate.EndInvoke(result);
            } 
            catch (HttpServerException)
            {
                //This would be thrown if we have an error processing the request
                //and will already have been logged
            }

            //Finish up the request
            try
            {
                if (this._listener.IsListening && context != null)
                {
                    this.ApplyPreResponseModules(context);

                    //Always flush and close the response stream before closing the response
                    context.Response.OutputStream.Flush();
                    context.Response.OutputStream.Close();
                    context.Response.Close();
                }
            }
            catch (Exception ex)
            {
                this.LogError("Error finishing a HTTP request", ex);
            }
        }

        /// <summary>
        /// Applies pre-request modules returning whether actual request handling should be skipped
        /// </summary>
        /// <param name="context">Server Context</param>
        /// <returns></returns>
        private bool ApplyPreRequestModules(HttpServerContext context)
        {
            foreach (IHttpListenerModule module in this._preRequestModules)
            {
                try
                {
                    //If a Module returns false then Handling should be skipped and go onto pre-response modules
                    if (!module.ProcessRequest(context)) return true;
                }
                catch (Exception ex)
                {
                    this.LogError("Error in pre-request module " + module.GetType().FullName, ex);
                }
            }

            //If all Modules return true then Handling continues normally
            return false;
        }

        /// <summary>
        /// Applies pre-response modules
        /// </summary>
        /// <param name="context">Server Context</param>
        private void ApplyPreResponseModules(HttpServerContext context)
        {
            foreach (IHttpListenerModule module in this._preResponseModules)
            {
                try
                {
                    //If a Module returns false then further pre-response modules are skipped
                    if (!module.ProcessRequest(context)) return;
                }
                catch (Exception ex)
                {
                    this.LogError("Error in pre-response module " + module.GetType().FullName, ex);
                }
            }
        }

        /// <summary>
        /// Maps a URL Path to a physical path returning null if no such mapping is possible
        /// </summary>
        /// <param name="path">URL Path</param>
        /// <returns>Either a Physical Path or null is not a valid path</returns>
        public String MapPath(String path)
        {
            String realPath;
            if (this._virtualDirs != null)
            {
                //Try to Map against Virtual Directories
                String relativePath;
                realPath = this._virtualDirs.GetDirectory(path, out relativePath);
                if (realPath != null)
                {
                    //Maps to a Virtual Directory so resolve the path relative to the virtual directory
                    //Not allowed to break out of directory, if the content you output has parent paths in it 
                    //then is up to the client to resolve these themselves before making subsequent requests
                    if (relativePath.StartsWith("/")) relativePath = relativePath.Substring(1);
                    if (relativePath.Contains("../") || relativePath.Contains("./")) return null;
                    realPath = Path.Combine(realPath, relativePath);
                    return realPath;
                }
            }
            
            //If Virtual Directories not available or don't apply then map against Base Directory
            if (this._baseDirectory != null)
            {
                if (path.StartsWith("/")) path = path.Substring(1);

                //Try to ensure that paths cannot go outside the base directory
                if (path.Contains("../") || path.Contains("./")) return null;
                realPath = Path.Combine(this._baseDirectory, path.Replace('/', Path.DirectorySeparatorChar));
                if (realPath.StartsWith(this._baseDirectory))
                {
                    return realPath;
                }
                else
                {
                    //If URL Path tries to go outside base directory then this should fail
                    return null;
                }
            }
            else
            {
                //No Base Directory so cannot map URL paths to physical paths
                return null;
            }
        }

        /// <summary>
        /// Gets the MIME Type for the response based on the file extension or null if the MIME type is not allowed to be served
        /// </summary>
        /// <param name="extension">File Extension</param>
        /// <returns></returns>
        public MimeTypeMapping GetMimeType(String extension)
        {
            return this._mimeTypes.GetMapping(extension);
        }

        /// <summary>
        /// Remaps the Request to another Handler, the Handler must be a registered Handler for remapping to succeed
        /// </summary>
        /// <param name="context">HTTP Context</param>
        /// <param name="handlerType">Handler Type</param>
        public void RemapHandler(HttpServerContext context, Type handlerType)
        {
            try
            {
                IHttpListenerHandler handler = this._handlers.GetHandler(handlerType);
                handler.ProcessRequest(context);
            }
            catch (NoHandlerException noHandlerEx)
            {
                this.LogError(noHandlerEx);
                context.Response.StatusCode = (int)HttpStatusCode.MethodNotAllowed;
            }
            catch (HttpServerException serverEx)
            {
                this.LogError(serverEx);
                context.Response.StatusCode = (int)HttpStatusCode.InternalServerError;
            }
        }

        /// <summary>
        /// Log an error
        /// </summary>
        /// <param name="message">Message</param>
        /// <param name="ex">Exception</param>
        internal void LogError(String message, Exception ex)
        {
            this._console.Error(message, ex);
        }

        /// <summary>
        /// Log an error
        /// </summary>
        /// <param name="ex">Error</param>
        internal void LogError(Exception ex)
        {
            this._console.Error(ex);
        }

        #endregion

        /// <summary>
        /// Disposes of the Server, once disposed of a Server instance cannot be reused and its methods will throw <see cref="ObjectDisposedException"/> if this is attempted
        /// </summary>
        public void Dispose()
        {
            this.DisposeInternal(true);
        }

        /// <summary>
        /// Actually dispose of the Server
        /// </summary>
        /// <param name="disposing">Whether this was called from the Dispose() method or not</param>
        private void DisposeInternal(bool disposing)
        {
            try
            {
                Monitor.Enter(this);
                if (this._disposed) return;
                if (disposing) GC.SuppressFinalize(this);

                //Stop if we are currently running
                if (this.IsRunning)
                {
                    this.Stop();
                }

                //Tell stuff we are being disposed of
                this.PreStateChange(ServerState.Disposing);
                this.State = ServerState.Disposing;
                this.PostStateChange();

                //Do the actual disposal
                this.PreStateChange(ServerState.Disposed);
                if (this._listener != null)
                {
                    if (this._listener.IsListening) this._listener.Stop();
                }
                this.State = ServerState.Disposed;
                this.PostStateChange();

                //Dispose of server components
                this._handlers.Dispose();
                foreach (IHttpListenerModule module in this._preRequestModules)
                {
                    module.Dispose();
                }
                foreach (IHttpListenerModule module in this._preResponseModules)
                {
                    module.Dispose();
                }
                foreach (IHttpLogger logger in this._loggers)
                {
                    logger.Dispose();
                }
                this._console.Dispose();

                //Now mark ourselves as disposed
                this._disposed = true;

                //Wait for termination
                this._shouldTerminate = true;
                while (this._serverThread.ThreadState != System.Threading.ThreadState.Stopped)
                {
                    Thread.Sleep(25);
                }
            }
            finally
            {
                Monitor.Exit(this);
            }
        }
    }
}
