/*
VDS.Web.Server is licensed under the MIT License

Copyright (c) 2010-2013 Robert Vesse

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using VDS.Web.Handlers;
using VDS.Web.Logging;

namespace VDS.Web
{
    [TestClass]
    public class ComponentStateTests
    {
        private readonly int TestPort = PortManager.GetNextPort();

        private HttpServer _server;

        [TestInitialize]
        public void Setup()
        {
            HttpListenerHandlerCollection handlers = new HttpListenerHandlerCollection();
            handlers.AddMapping(new HttpRequestMapping(HttpRequestMapping.AllVerbs, HttpRequestMapping.AnyPath, typeof(RequestInfoHandler)));
            this._server = new HttpServer(TestPort, handlers);
            this._server.Start();
        }

        [TestCleanup]
        public void Teardown()
        {
            this._server.Dispose();
        }

        private String ServerUrl
        {
            get
            {
                return "http://localhost:" + TestPort + "/test";
            }
        }

        [TestMethod]
        public void ComponentStateOverStop1()
        {
            RequestInfoHandler handler = null;
            using (WebClient client = new WebClient())
            {
                client.DownloadString(this.ServerUrl);

                handler = this._server.Handlers.GetHandler(typeof(RequestInfoHandler)) as RequestInfoHandler;
                Assert.IsNotNull(handler);
                Assert.IsNotNull(handler.Context);
                Assert.IsTrue(handler.Form.IsValid);
                Assert.IsTrue(handler.Form.IsEmpty);
            }

            //Stopping the server should discard state e.g. cached handlers
            this._server.Stop();

            handler = null;
            handler = this._server.Handlers.GetHandler(typeof(RequestInfoHandler)) as RequestInfoHandler;
            Assert.IsNotNull(handler);
            Assert.IsNull(handler.Context);
        }

        [TestMethod]
        public void ComponentStateOverStop2()
        {
            HttpServer server = null;
            try
            {
                HttpListenerHandlerCollection handlers = new HttpListenerHandlerCollection();
                handlers.AddMapping(new HttpRequestMapping(HttpRequestMapping.AllVerbs, HttpRequestMapping.AnyPath, typeof(RequestInfoHandler)));
                server = new HttpServer(PortManager.GetNextPort(), handlers);

                FileLogger logger = new FileLogger("test.log");
                server.AddLogger(logger);

                server.Start();

                //Make a request
                using (WebClient client = new WebClient())
                {
                    client.DownloadString(this.ServerUrl);

                    RequestInfoHandler handler = this._server.Handlers.GetHandler(typeof(RequestInfoHandler)) as RequestInfoHandler;
                    Assert.IsNotNull(handler);
                    Assert.IsNotNull(handler.Context);
                    Assert.IsTrue(handler.Form.IsValid);
                    Assert.IsTrue(handler.Form.IsEmpty);
                }

                //Now stop the server
                server.Stop();

                //After the server is stopped we should be able to access the file that the logger was holding open
                FileStream stream = File.OpenRead("test.log");
                stream.Close();

                //Restarting the server should reacquire the file
                server.Start();

                try
                {
                    stream = File.OpenRead("test.log");
                    Assert.Fail("The server should be holding test.log open so we should not be able to access it");
                }
                catch
                {
                    //Ignore - we expected this to fail
                }

                server.Dispose();
                server = null;

                //After the server is disposed we should be able to access the file that the logger was holding open again
                stream = File.OpenRead("test.log");
                stream.Close();
            }
            finally
            {
                if (server != null)
                {
                    server.Dispose();
                }
            }
        }

        [TestMethod]
        public void ComponentStateOverDispose1()
        {
            HttpServer server = null;
            try
            {
                HttpListenerHandlerCollection handlers = new HttpListenerHandlerCollection();
                handlers.AddMapping(new HttpRequestMapping(HttpRequestMapping.AllVerbs, HttpRequestMapping.AnyPath, typeof(RequestInfoHandler)));
                server = new HttpServer(PortManager.GetNextPort(), handlers);

                FileLogger logger = new FileLogger("test.log");
                server.AddLogger(logger);

                server.Start();

                //Make a request
                using (WebClient client = new WebClient())
                {
                    client.DownloadString(this.ServerUrl);

                    RequestInfoHandler handler = this._server.Handlers.GetHandler(typeof(RequestInfoHandler)) as RequestInfoHandler;
                    Assert.IsNotNull(handler);
                    Assert.IsNotNull(handler.Context);
                    Assert.IsTrue(handler.Form.IsValid);
                    Assert.IsTrue(handler.Form.IsEmpty);
                }

                //Now dipose of the server
                server.Dispose();
                server = null;

                //After the server is disposed we should be able to access the file that the logger was holding open
                FileStream stream = File.OpenRead("test.log");
                stream.Close();
            }
            finally
            {
                if (server != null)
                {
                    server.Dispose();
                }
            }
        }
    }
}
