/*
VDS.Web.Server is licensed under the MIT License

Copyright (c) 2010-2013 Robert Vesse

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using VDS.Web.Handlers;

namespace VDS.Web
{
    [TestClass]
    public class FormParsingTest
    {
        private readonly int TestPort = PortManager.GetNextPort();

        private HttpServer _server;
        private RequestInfoHandler _handler;

        [TestInitialize]
        public void Setup()
        {
            HttpListenerHandlerCollection handlers = new HttpListenerHandlerCollection();
            handlers.AddMapping(new HttpRequestMapping(HttpRequestMapping.AllVerbs, HttpRequestMapping.AnyPath, typeof(RequestInfoHandler)));
            this._server = new HttpServer(TestPort, handlers);
            this._server.Start();

            this._handler = this._server.Handlers.GetHandler(typeof(RequestInfoHandler)) as RequestInfoHandler;

            Assert.IsNotNull(this._handler);
        }

        [TestCleanup]
        public void Teardown()
        {
            this._server.Dispose();
        }

        private String ServerUrl
        {
            get
            {
                return "http://localhost:" + TestPort + "/test";
            }
        }

        [TestMethod]
        public void FormParsing01()
        {
            try
            {
                using (WebClient client = new WebClient())
                {
                    client.DownloadString(this.ServerUrl);

                    Assert.IsNotNull(this._handler.Context);
                    Assert.IsTrue(this._handler.Form.IsValid);
                    Assert.IsTrue(this._handler.Form.IsEmpty);
                }
            }
            finally
            {
                this._handler.Reset();
            }
        }

        [TestMethod]
        public void FormParsing02()
        {
            try
            {
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(this.ServerUrl);
                request.Method = "POST";
                request.ContentType = FormVariables.WWWFormURLEncoded;

                using (StreamWriter writer = new StreamWriter(request.GetRequestStream()))
                {
                    writer.Write("key=value");
                }

                using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                {
                    response.Close();

                    Assert.IsNotNull(this._handler.Context);
                    Assert.IsTrue(this._handler.Form.IsValid);
                    Assert.IsFalse(this._handler.Form.IsEmpty);
                    Assert.AreEqual(1, this._handler.Form.Count);
                    Assert.AreEqual("value", this._handler.Form["key"]);
                }
            }
            finally
            {
                this._handler.Reset();
            }
        }

        [TestMethod]
        public void FormParsing03()
        {
            try
            {
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(this.ServerUrl);
                request.Method = "POST";
                request.ContentType = FormVariables.WWWFormURLEncoded;

                using (StreamWriter writer = new StreamWriter(request.GetRequestStream()))
                {
                    writer.Write("key=value&key2=value2");
                }

                using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                {
                    response.Close();

                    Assert.IsNotNull(this._handler.Context);
                    Assert.IsTrue(this._handler.Form.IsValid);
                    Assert.IsFalse(this._handler.Form.IsEmpty);
                    Assert.AreEqual(2, this._handler.Form.Count);
                    Assert.AreEqual("value", this._handler.Form["key"]);
                    Assert.AreEqual("value2", this._handler.Form["key2"]);
                }
            }
            finally
            {
                this._handler.Reset();
            }
        }

        [TestMethod]
        public void FormParsing04()
        {
            try
            {
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(this.ServerUrl);
                request.Method = "POST";
                request.ContentType = FormVariables.WWWFormURLEncoded;

                using (StreamWriter writer = new StreamWriter(request.GetRequestStream()))
                {
                    writer.Write("key=some value");
                }

                using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                {
                    response.Close();

                    Assert.IsNotNull(this._handler.Context);
                    Assert.IsTrue(this._handler.Form.IsValid);
                    Assert.IsFalse(this._handler.Form.IsEmpty);
                    Assert.AreEqual(1, this._handler.Form.Count);
                    Assert.AreEqual("some value", this._handler.Form["key"]);
                }
            }
            finally
            {
                this._handler.Reset();
            }
        }

        [TestMethod]
        public void FormParsing05()
        {
            try
            {
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(this.ServerUrl);
                request.Method = "POST";
                request.ContentType = FormVariables.WWWFormURLEncoded;

                using (StreamWriter writer = new StreamWriter(request.GetRequestStream()))
                {
                    writer.Write("key=" + Uri.EscapeDataString("some value"));
                }

                using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                {
                    response.Close();

                    Assert.IsNotNull(this._handler.Context);
                    Assert.IsTrue(this._handler.Form.IsValid);
                    Assert.IsFalse(this._handler.Form.IsEmpty);
                    Assert.AreEqual(1, this._handler.Form.Count);
                    Assert.AreEqual("some value", this._handler.Form["key"]);
                }
            }
            finally
            {
                this._handler.Reset();
            }
        }

        [TestMethod]
        public void FormParsing06()
        {
            try
            {
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(this.ServerUrl);
                request.Method = "POST";
                request.ContentType = FormVariables.WWWFormURLEncoded;

                using (StreamWriter writer = new StreamWriter(request.GetRequestStream()))
                {
                    writer.Write("key=some+value");
                }

                using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                {
                    response.Close();

                    Assert.IsNotNull(this._handler.Context);
                    Assert.IsTrue(this._handler.Form.IsValid);
                    Assert.IsFalse(this._handler.Form.IsEmpty);
                    Assert.AreEqual(1, this._handler.Form.Count);
                    Assert.AreEqual("some value", this._handler.Form["key"]);
                }
            }
            finally
            {
                this._handler.Reset();
            }
        }

        [TestMethod]
        public void FormParsing07()
        {
            try
            {
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(this.ServerUrl);
                request.Method = "POST";
                request.ContentType = FormVariables.WWWFormURLEncoded;

                using (StreamWriter writer = new StreamWriter(request.GetRequestStream()))
                {
                    writer.Write("key=some\nvalue");
                }

                using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                {
                    response.Close();

                    Assert.IsNotNull(this._handler.Context);
                    Assert.IsTrue(this._handler.Form.IsValid);
                    Assert.IsFalse(this._handler.Form.IsEmpty);
                    Assert.AreEqual(1, this._handler.Form.Count);
                    Assert.AreEqual("some\nvalue", this._handler.Form["key"]);
                }
            }
            finally
            {
                this._handler.Reset();
            }
        }

         [TestMethod]
        public void FormParsing08()
        {
            try
            {
                HttpWebRequest request = (HttpWebRequest)WebRequest.Create(this.ServerUrl);
                request.Method = "POST";
                request.ContentType = FormVariables.WWWFormURLEncoded;

                using (StreamWriter writer = new StreamWriter(request.GetRequestStream()))
                {
                    writer.Write("key=value;key2=value2");
                }

                using (HttpWebResponse response = (HttpWebResponse)request.GetResponse())
                {
                    response.Close();

                    Assert.IsNotNull(this._handler.Context);
                    Assert.IsTrue(this._handler.Form.IsValid);
                    Assert.IsFalse(this._handler.Form.IsEmpty);
                    Assert.AreEqual(2, this._handler.Form.Count);
                    Assert.AreEqual("value", this._handler.Form["key"]);
                    Assert.AreEqual("value2", this._handler.Form["key2"]);
                }
            }
            finally
            {
                this._handler.Reset();
            }
        }

        [TestMethod]
        public void UriEncoding01()
        {
            String encoded = Uri.EscapeDataString("some value");
            Assert.AreEqual("some%20value", encoded);
        }

        [TestMethod]
        public void UriEncoding02()
        {
            String encoded = Uri.EscapeDataString("+");
            Assert.AreEqual("%2B", encoded);
        }

    }
}
