/*
VDS.Web.Server is licensed under the MIT License

Copyright (c) 2010-2013 Robert Vesse

Permission is hereby granted, free of charge, to any person obtaining a copy of this software
and associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software 
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using VDS.Web.Consoles;
using VDS.Web.Handlers;

namespace VDS.Web
{
    [TestClass]
    public class ServerControlTests
    {
        private readonly int TestPort = PortManager.GetNextPort();

        [TestMethod]
        public void ServerStartStop1()
        {
            //Simple Start, Stop, Dispose
            HttpServer server = new HttpServer(TestPort, new StaticFileHandlersCollection());
            Assert.AreEqual(ServerState.Created, server.State);
            Assert.IsFalse(server.IsRunning);
            server.Start();
            Assert.AreEqual(ServerState.Running, server.State);
            Assert.IsTrue(server.IsRunning);
            server.Stop();
            Assert.AreEqual(ServerState.Stopped, server.State);
            Assert.IsFalse(server.IsRunning);
            server.Dispose();
            Assert.AreEqual(ServerState.Disposed, server.State);
        }

        [TestMethod, ExpectedException(typeof(ObjectDisposedException))]
        public void ServerStartStop2()
        {
            //Simple Start, Stop, Dispose
            HttpServer server = new HttpServer(TestPort, new StaticFileHandlersCollection());
            Assert.AreEqual(ServerState.Created, server.State);
            Assert.IsFalse(server.IsRunning);
            server.Start();
            Assert.AreEqual(ServerState.Running, server.State);
            Assert.IsTrue(server.IsRunning);
            server.Stop();
            Assert.AreEqual(ServerState.Stopped, server.State);
            Assert.IsFalse(server.IsRunning);
            server.Dispose();
            Assert.AreEqual(ServerState.Disposed, server.State);

            //Should cause an ObjectDisposedException
            int port = server.Port;
        }

        [TestMethod]
        public void ServerStartStop3()
        {
            //Simple Start, Stop, Dipose
            //With checking of the console
            HttpServer server = new HttpServer(TestPort, new StaticFileHandlersCollection());
            Assert.AreEqual(ServerState.Created, server.State);
            Assert.IsFalse(server.IsRunning);

            StringWriter console = new StringWriter();
            server.Console = new StreamConsole(console);

            server.Start();
            Assert.AreEqual(ServerState.Running, server.State);
            Assert.IsTrue(server.IsRunning);
            Assert.IsTrue(console.ToString().Contains("Starting"));
            Assert.IsTrue(console.ToString().Contains("Started"));

            server.Stop();
            Assert.AreEqual(ServerState.Stopped, server.State);
            Assert.IsFalse(server.IsRunning);
            Assert.IsTrue(console.ToString().Contains("Stopping"));
            Assert.IsTrue(console.ToString().Contains("Stopped"));

            server.Dispose();
            Assert.AreEqual(ServerState.Disposed, server.State);
        }

        [TestMethod, ExpectedException(typeof(HttpServerException))]
        public void ServerInvalidStart1()
        {
            HttpServer server = new HttpServer(TestPort, new StaticFileHandlersCollection());
            try
            {
                Assert.AreEqual(ServerState.Created, server.State);
                Assert.IsFalse(server.IsRunning);
                server.Start();
                Assert.AreEqual(ServerState.Running, server.State);
                Assert.IsTrue(server.IsRunning);

                //Calling Start() again should be invalid
                server.Start();
            }
            finally
            {
                server.Dispose();
            }
        }

        [TestMethod, ExpectedException(typeof(HttpServerException))]
        public void ServerInvalidStop1()
        {
            HttpServer server = new HttpServer(TestPort, new StaticFileHandlersCollection());
            try
            {
                Assert.AreEqual(ServerState.Created, server.State);
                Assert.IsFalse(server.IsRunning);

                //Calling Stop() should be invalid when the server is not running
                server.Stop();
            }
            finally
            {
                server.Dispose();
            }
        }
    }
}
